#include <pthread.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#define CHECK(op, msg) { if ((errno = (op)) != 0) { perror(msg); exit(1); } }

struct child
{
    pthread_barrier_t *b;
    int num;
};

void *child(void *arg)
{
    struct child *c = (struct child*) arg;
    pthread_barrier_wait(c->b);
    pause();
    printf("%d\n", c->num);
    fflush(stdout);
    pthread_exit(NULL);
}

void onsig(int signum)
{
    (void) signum;
}

int main(int argc, char const *argv[])
{
    if (argc != 2)
    {
        fprintf(stderr, "bad usage: %s <n>\n", argv[0]);
        return 1;
    }

    int n = atoi(argv[1]);
    if (n < 0)
    {
        fprintf(stderr, "error: %d < 0\n", n);
        return 1;
    } else if (n == 0)
    {
        return 0;
    }
    

    pthread_barrier_t b;
    // pthread_barrierattr_t ba;

    // pthread_barrierattr_init(&ba);
    // pthread_barrierattr_setpshared (&ba ,PTHREAD_PROCESS_SHARED);
    CHECK(pthread_barrier_init(&b, NULL, n), "init barrier");

    signal(SIGUSR1, onsig);
    signal(SIGINT, onsig);
    
    pthread_t *threads = malloc(n*sizeof(pthread_t));
    struct child *args = malloc(n*sizeof(struct child));

    for (size_t i = 0; i < n; i++)
    {
        args[i].b = &b;
        args[i].num = i;
        CHECK(pthread_create(threads+i, NULL, child, args+i), "thread create");
    }

    sigset_t sigs;
    sigemptyset(&sigs);
    sigaddset(&sigs, SIGINT);
    
    for (size_t i = 0; i < n; i++)
    {
        int sig;
        sigwait(&sigs, &sig);
        CHECK(pthread_kill(threads[i], SIGUSR1), "send SIGUSR1");
        CHECK(pthread_join(threads[i], NULL), "join");
    }
    

    // pthread_barrierattr_destroy(&ba);

    free(threads);
    free(args);

    return 0;
}
