#!/bin/sh

PROG=${PROG:=./exo2}

DELAI_INIT=0.1			# 100 ms d'attente pour l'initialisation
DELAI=0.02			# 20 ms d'attente, devraient être suffisantes

#
# Script Shell de test de l'exercice 2
# Utilisation : sh ./test2.sh
#
# Si tout se passe bien, le script doit afficher "Tests ok" à la fin
# Dans le cas contraire, le nom du test échoué s'affiche.
# Les fichiers sont laissés dans /tmp/test2*, vous pouvez les examiner
# Pour avoir plus de détails sur l'exécution du script, vous pouvez
# utiliser :
#	sh -x ./test2.sh
# Toutes les commandes exécutées par le script sont alors affichées.
#

TMP=/tmp/test2

# il ne faudrait jamais appeler cette fonction
# argument : message d'erreur
fail ()
{
    local msg="$1"

    echo "Échec du test '$msg'."
    exit 1
}

# Si le processus est terminé, signaler une erreur
# $1 = pid
# $2 = msg
ps_termine ()
{
    [ $# != 2 ] && fail "ERREUR SYNTAXE ps_termine"
    local pid="$1" msg="$2"
    kill -0 $pid 2> /dev/null || fail "$msg"
}

# Si le processus existe toujours, signaler une erreur
# $1 = pid
# $2 = msg
ps_existe ()
{
    [ $# != 2 ] && fail "ERREUR SYNTAXE ps_existe"
    local pid="$1" msg="$2"
    kill -0 $pid 2> /dev/null && fail "$msg"
}

est_vide ()
{
    local fichier="$1"
    test $(wc -l < "$fichier") = 0
}

# Vérifie que le message d'erreur est envoyé sur la sortie d'erreur
# et non sur stdout
verifier_stderr ()
{
    [ $# != 1 ] && fail "ERREUR SYNTAXE verifier_stderr"
    local base="$1"
    est_vide $base.err \
	&& fail "Le message d'erreur devrait être sur la sortie d'erreur"
    est_vide $base.out \
	|| fail "Rien ne devrait être affiché sur la sortie standard"
}

# Vérifie que le fils s'est bien terminé par un code de retour nul
# $1 = pid à vérifier
# $2 = msg
wait_ok ()
{
    [ $# != 2 ] && fail "ERREUR SYNTAXE wait_ok"
    local pid="$1" msg="$2"
    wait $pid || fail "$msg"
}

# Comparer la sortie standard avec l'ordre "normal"
# $1 = max
# $2 = fichier (sortie standard de la commande)
# $3 = msg
comparer_stdout ()
{
    [ $# != 3 ] && fail "ERREUR SYNTAXE comparer_stdout"
    local max="$1" file="$2" msg="$3"
    seq 0 "$max" | diff -q "$file" -	|| fail "$msg"
}

# Tester pour N threads fils
# $1 = n
tester_n ()
{
    [ $# != 1 ] && fail "ERREUR SYNTAXE tester_n"

    local n=$1
    $PROG $n  > $TMP.out 2> $TMP.err &
    PID=$!

    sleep $DELAI_INIT
    for i in $(seq 0 $((n-1)))
    do
	ps_termine "$PID" "n=$n, processus terminé prématurément après $i ^C"
	comparer_stdout $((i-1)) $TMP.out "n=$n, sortie invalide après $i ^C"
	kill -INT $PID
	sleep $DELAI
    done
    ps_existe $PID "n=$n, processus non terminé après $n ^C"
    wait_ok $PID "n=$n, code de retour != 0"
    comparer_stdout $((n-1)) $TMP.out "n=$n, sortie finale invalide"
}

##############################################################################
# Tests d'erreur sur les arguments

# Erreur sur le nombre d'arguments
$PROG    > $TMP.out 2> $TMP.err		&& fail "pas d'arg"
verifier_stderr $TMP

# Erreur sur le nombre d'arguments
$PROG 1 2 > $TMP.out 2> $TMP.err	&& fail "2 args"
verifier_stderr $TMP

# Erreur sur la valeur de l'argument
$PROG -1  > $TMP.out 2> $TMP.err	&& fail "n = -1"
verifier_stderr $TMP

##############################################################################
# Avec n = 0, le thread père ne devrait générer aucun fils et se terminer
$PROG 0  > $TMP.out 2> $TMP.err &
PID=$!

sleep $DELAI_INIT
ps_existe $PID "n=0, processus non terminé"
wait_ok $PID "n=0, code de retour != 0"

##############################################################################
# Avec n = 1, on devrait sortir dès le premier ^C
$PROG 1  > $TMP.out 2> $TMP.err &
PID=$!

sleep $DELAI_INIT
ps_termine $PID "n=1, processus terminé prématurément"
kill -INT $PID
sleep $DELAI
ps_existe $PID "n=1, processus non terminé après 1 ^C"
wait_ok $PID "n=1, code de retour != 0"
comparer_stdout 0 $TMP.out "n=1, affichage invalide"

##############################################################################
# Avec n = 5, les résultats doivent être testés au fur et à mesure
tester_n 5

##############################################################################
# Avec n = 100, les résultats doivent être testés au fur et à mesure
# Note : avec des valeurs de "n" plus importantes, les délais ne suffisent
# plus pour couvrir les temps de création et de gestion des threads. Les
# augmenter serait possible, mais la durée du script serait alors trop grande.
tester_n 100

##############################################################################
# Fini !

rm -f $TMP.*

echo "Tests ok"
exit 0
