#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdnoreturn.h>
#include <semaphore.h>

#define MAXLEN          1000    // longueur maximum d'une ligne

#define CHK(op)   do { if ((op) == -1) { raler (1, #op); } } while (0)
#define NCHK(op)   do { if ((op) == NULL) { raler (1, #op); } } while (0)
#define TCHK(op)   do { if ((errno=(op)) > 0) { raler (1, #op); } } while (0)

struct arg
{
    char *fichier ;             // fichier que le thread doit lire
    char ligne [MAXLEN] ;       // chaque ligne lue par le thread
    pthread_mutex_t verrou;
    sem_t seml;
    sem_t seme;
    int termine;
} ;

noreturn void raler (int syserr, const char *fmt, ...)
{
    va_list ap ;

    va_start (ap, fmt) ;
    vfprintf (stderr, fmt, ap) ;
    fprintf (stderr, "\n") ;
    va_end (ap) ;

    if (syserr)
        perror ("") ;

    exit (EXIT_FAILURE) ;
}
 

void *f (void *arg)
{
    struct arg *a = arg ;
    char ligne [MAXLEN] ;               // y compris le \0 de fin de chaîne
    FILE *fp ;

    NCHK (fp = fopen (a->fichier, "r")) ;
    // lire la ligne dans la variable locale du thread, pendant que le thread
    // principal affiche la ligne partagée
    while (fgets (ligne, sizeof ligne, fp) != NULL)
    {
        CHK(sem_wait(&a->seme));
        // recopier la ligne lue dans la variable partagée 
        TCHK(pthread_mutex_lock(&a->verrou));
        strcpy (a->ligne, ligne) ;
        TCHK(pthread_mutex_unlock(&a->verrou));
        CHK(sem_post(&a->seml));
    }
    a->termine = 1;
    CHK(sem_post(&a->seml));
    if (fclose (fp) == EOF)
        raler (1, "fclose %s", a->fichier) ;

    return NULL ;
}

int main (int argc, char *argv [])
{
    pthread_t id ;
    struct arg arg ;
    int termine ;

    if (argc != 2)
        raler (0, "usage: %s fichier", argv [0]) ;

    arg.fichier = argv [1] ;
    arg.termine = 0;

    TCHK(pthread_mutex_init(&arg.verrou, NULL));
    CHK(sem_init(&arg.seml, 1, 0));
    CHK(sem_init(&arg.seme, 1, 1));
    // struct timespec t = {
    //     .tv_sec = 0, .tv_nsec = 5000
    // };
    // TCHK(pthread_mutex_timedlock(&arg.verrou, &t));
    TCHK (pthread_create (&id, NULL, f, &arg)) ;

    termine = 0 ;
    while (! termine)
    {
        CHK(sem_wait(&arg.seml));
        if (arg.termine)
            break;
        
        TCHK(pthread_mutex_lock(&arg.verrou));
        printf ("%s", arg.ligne) ;              // arg.ligne contient déjà \n
        termine = arg.termine;
        TCHK(pthread_mutex_unlock(&arg.verrou));
        CHK(sem_post(&arg.seme));
    }

    TCHK (pthread_join (id, NULL)) ;
    CHK(sem_destroy(&arg.seml));
    CHK(sem_destroy(&arg.seme));

    TCHK(pthread_mutex_destroy(&arg.verrou));

    exit (0) ;
}
