#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdnoreturn.h>
#include <semaphore.h>

#define MAXLEN          1000    // longueur maximum d'une ligne

#define NCHK(op)   do { if ((op) == NULL) { raler (1, #op); } } while (0)
#define TCHK(op)   do { if ((errno=(op)) > 0) { raler (1, #op); } } while (0)

struct arg
{
    char *fichier ;             // fichier que le thread doit lire
    char ligne [MAXLEN] ;       // chaque ligne lue par le thread

    sem_t ecrire;
    sem_t lire;
} ;

noreturn void raler (int syserr, const char *fmt, ...)
{
    va_list ap ;

    va_start (ap, fmt) ;
    vfprintf (stderr, fmt, ap) ;
    fprintf (stderr, "\n") ;
    va_end (ap) ;

    if (syserr)
        perror ("") ;

    exit (EXIT_FAILURE) ;
}
 

void *f (void *arg)
{
    struct arg *a = arg ;
    char ligne [MAXLEN] ;               // y compris le \0 de fin de chaîne
    FILE *fp ;

    NCHK (fp = fopen (a->fichier, "r")) ;
    // lire la ligne dans la variable locale du thread, pendant que le thread
    // principal affiche la ligne partagée
    while (fgets (ligne, sizeof ligne, fp) != NULL)
    {
        TCHK(sem_wait(&a->ecrire));
        // recopier la ligne lue dans la variable partagée 
        strcpy (a->ligne, ligne) ;
        TCHK(sem_post(&a->lire));
    }
    a->fichier = NULL;
    if (fclose (fp) == EOF)
        raler (1, "fclose %s", a->fichier) ;

    return NULL ;
}

int main (int argc, char *argv [])
{
    pthread_t id ;
    struct arg arg ;
    int termine ;

    if (argc != 2)
        raler (0, "usage: %s fichier", argv [0]) ;

    arg.fichier = argv [1] ;

    TCHK(sem_init(&arg.ecrire,1, 0));
    TCHK(sem_init(&arg.lire,1,0));

    TCHK (pthread_create (&id, NULL, f, &arg)) ;

    TCHK(sem_post(&arg.ecrire));
    termine = 0 ;
    while (! termine)
    {
        TCHK(sem_wait(&arg.lire));
        printf ("%s", arg.ligne) ;              // arg.ligne contient déjà \n
        
        termine = (arg.fichier == NULL);
        TCHK(sem_post(&arg.ecrire));
    }

    TCHK (pthread_join (id, NULL)) ;
    TCHK(sem_destroy(&arg.ecrire));
    TCHK(sem_destroy(&arg.lire));

    exit (0) ;
}
